@@ -12,6 +12,7 @@ import statistics
 import pkg_resources
 import decimal
 import chemparse
+import copy
 
 import rpy2.rinterface_lib.callbacks
 import logging
@@ -791,7 +792,7 @@ def animation(basis_args={}, species_args={}, affinity_args={},
     fig.show(config=config)
 
     
-def diagram_interactive(data, title=None,
+def diagram_interactive(data, title=None, borders=0,
                         annotation=None, annotation_coords=[0, 0],
                         balance=None, xlab=None, ylab=None, colormap="viridis",
                         width=600, height=520, alpha=False, messages=True,
@@ -808,6 +809,11 @@ def diagram_interactive(data, title=None,
     title : str, optional
         Title of the plot.
     
+    borders : float, default 0
+        If set to a value greater than 0, shows lines that indicate borders
+        between regions in predominance diagrams. Value indicates thickness
+        of the border, in pixels.
+    
     annotation : str, optional
         Annotation to add to the plot.
     
@@ -857,10 +863,10 @@ def diagram_interactive(data, title=None,
     An interactive plot.
     """
     
-    basis_sp = list(data.rx2("basis").index)
+    basis_sp = list(ro.conversion.rpy2py(data.rx2("basis")).index)
 
-    xyvars = list(data.rx2("vars"))
-    xyvals = list(data.rx2("vals"))
+    xyvars = list(ro.conversion.rpy2py(data.rx2("vars")))
+    xyvals = list(ro.conversion.rpy2py(data.rx2("vals")))
 
     if 'loga.equil' not in data.names:
         calc_type = "a"
@@ -1013,6 +1019,9 @@ def diagram_interactive(data, title=None,
         fig.update(data=[{'customdata': dmap_names,
             'hovertemplate': xlab+': %{x}<br>'+ylab+': %{y}<br>Region: %{customdata}<extra></extra>'}])
 
+        if colormap == 'none':
+            colormap = [[0, 'white'], [1, 'white']]
+        
         fig.update_traces(dict(showscale=False, 
                                coloraxis=None, 
                                colorscale=colormap),
@@ -1060,10 +1069,100 @@ def diagram_interactive(data, title=None,
                                           },
                  }
         
+        if borders > 0:
+            
+            unique_x_vals = list(dict.fromkeys(df[xvar]))
+            unique_y_vals = list(dict.fromkeys(df[yvar]))
+            
+            def mov_mean(numbers=[], window_size=2):
+                i = 0
+                moving_averages = []
+                while i < len(numbers) - window_size + 1:
+                    this_window = numbers[i : i + window_size]
+
+                    window_average = sum(this_window) / window_size
+                    moving_averages.append(window_average)
+                    i += 1
+                return moving_averages
+            
+            x_mov_mean = mov_mean(unique_x_vals)
+            y_mov_mean = mov_mean(unique_y_vals)
+
+            x_plot_min = x_mov_mean[0] - (x_mov_mean[1] - x_mov_mean[0])
+            y_plot_min = y_mov_mean[0] - (y_mov_mean[1] - y_mov_mean[0])
+
+            x_plot_max = x_mov_mean[-1] + (x_mov_mean[1] - x_mov_mean[0])
+            y_plot_max = y_mov_mean[-1] + (y_mov_mean[1] - y_mov_mean[0])
+
+            x_vals_border = [x_plot_min] + x_mov_mean + [x_plot_max]
+            y_vals_border = [y_plot_min] + y_mov_mean + [y_plot_max]
+            
+            data = np.array(df.pred)
+            shape = (len(xvals), len(yvals))
+            dmap = data.reshape(shape)
+            
+            def find_line(dmap, row_index):
+                return [i for i in range(0, len(dmap[row_index])-1) if dmap[row_index][i] != dmap[row_index][i+1]]
+
+            nrows, ncols = dmap.shape
+            vlines = []
+            for row_i in range(0, nrows):
+                vlines.append(find_line(dmap, row_i))
+
+            dmap_transposed = dmap.transpose()
+            nrows, ncols = dmap_transposed.shape
+            hlines = []
+            for row_i in range(0, nrows):
+                hlines.append(find_line(dmap_transposed, row_i))
+            y_coord_list_vertical = []
+            x_coord_list_vertical = []
+            for i,row in enumerate(vlines):
+                for line in row:
+                    x_coord_list_vertical += [x_vals_border[line+1], x_vals_border[line+1], np.nan]
+                    y_coord_list_vertical += [y_vals_border[i], y_vals_border[i+1], np.nan]
+
+            y_coord_list_horizontal = []
+            x_coord_list_horizontal = []
+            for i,col in enumerate(hlines):
+                for line in col:
+                    y_coord_list_horizontal += [y_vals_border[line+1], y_vals_border[line+1], np.nan]
+                    x_coord_list_horizontal += [x_vals_border[i], x_vals_border[i+1], np.nan]
+                    
+            fig.add_trace(
+                go.Scatter(
+              mode= 'lines',
+              x= x_coord_list_horizontal,
+              y= y_coord_list_horizontal,
+              line= {
+                "width": borders,
+                "color": 'black'
+              },
+              hoverinfo= 'skip',
+              showlegend=False,
+                )
+            )
+            fig.add_trace(
+                go.Scatter(
+              mode= 'lines',
+              x= x_coord_list_vertical,
+              y= y_coord_list_vertical,
+              line= {
+                "width": borders,
+                "color": 'black'
+              },
+              hoverinfo= 'skip',
+              showlegend=False,
+                )
+            )
+        
+            fig.update_yaxes(range=[min(yvals), max(yvals)], autorange=False, mirror=True)
+            fig.update_xaxes(range=[min(xvals), max(xvals)], autorange=False, mirror=True)
+        
+        
     if plot_it:
         fig.show(config=config)
     
-    return df
+    return df, fig
 
 
 def _convert_to_RVector(value, force_Rvec=True):
@@ -1098,7 +1197,7 @@ def _convert_to_RVector(value, force_Rvec=True):
         return ro.BoolVector(value)
     elif all(isinstance(x, (int, np.integer)) for x in value):
         return ro.IntVector(value)
-    elif all(isinstance(x, (int, np.integer, float, np.float)) for x in value):
+    elif all(isinstance(x, (int, np.integer, float)) for x in value):
         return ro.FloatVector(value)
     else:
         return ro.StrVector(value)
@@ -1489,7 +1588,7 @@ def diagram(eout, ptype='auto', alpha=False, normalize=False,
             annotation=None, annotation_coords=[0,0],
             width=600, height=520, dpi=150,
             messages=True, interactive=False, save_as=None, save_format=None,
-            save_scale=1):
+            save_scale=1, fig_out=False):
     
     """
     Python wrapper for the diagram() function in CHNOSZ.
@@ -1680,7 +1779,10 @@ def diagram(eout, ptype='auto', alpha=False, normalize=False,
     save_scale : numeric, default 1
         For interactive plots only (`interactive`=True). Multiply
         title/legend/axis/canvas sizes by this factor when saving the figure.
-    
+        
+    fig_out : bool, default False
+        Function output is a plotly figure? Ignored if `interactive` is False.
+        
     Returns
     -------
     a : rpy2.ListVector
@@ -1689,8 +1791,15 @@ def diagram(eout, ptype='auto', alpha=False, normalize=False,
         Dictionary of arguments supplied to `diagram`.
     """
     
+    if lwd == None:
+        borders = 0
+    else:
+        borders = lwd
+    
     if interactive:
-        df = diagram_interactive(data=eout, title=main, annotation=annotation,
+            
+        df, fig = diagram_interactive(data=eout, title=main, borders=borders,
+                                 annotation=annotation,
                                  annotation_coords=annotation_coords,
                                  balance=balance,
                                  xlab=xlab, ylab=ylab,
@@ -1699,7 +1808,10 @@ def diagram(eout, ptype='auto', alpha=False, normalize=False,
                                  alpha=alpha, plot_it=plot_it,
                                  save_as=save_as, save_format=save_format,
                                  save_scale=save_scale, messages=messages)
-        return df
+        if fig_out:
+            return df, fig
+        else:
+            return df
     
     
     args = {'eout':eout, 'type':ptype, 'alpha':alpha, 'normalize':normalize,
@@ -2051,7 +2163,7 @@ def add_OBIGT(file, species=None, force=True, messages=True):
         elif ".csv" not in file[-4:] or ".CSV" in file[-4:]:
             raise Exception("File must be in .csv format")
         else:
-            df = pd.read_csv(file, keep_default_na=False) # keep_default_na=False keeps NA in data file instead of converting them to NaN. NaNs cause errors when converting to an R dataframe with rpy2 3.4.5.
+            df = pd.read_csv(file, keep_default_na=False, na_values=['NA']) # keep_default_na=False keeps NA in data file instead of converting them to NaN. NaNs cause errors when converting to an R dataframe with rpy2 3.4.5.
             if df.shape[0] == 0:
                 raise Exception("file is empty")
     elif isinstance(file, pd.DataFrame):
@@ -2070,7 +2182,7 @@ def add_OBIGT(file, species=None, force=True, messages=True):
 
     if all(col in df.columns for col in OBIGT_cols):
 
-        df_mod_OBIGT = df[OBIGT_cols]
+        df_mod_OBIGT = copy.deepcopy(df[OBIGT_cols])
 
         if species != None:
             if isinstance(species, list):
@@ -2089,6 +2201,12 @@ def add_OBIGT(file, species=None, force=True, messages=True):
         msg = "The file must contain all of the columns: {}".format(str(OBIGT_cols))
         raise Exception(msg)
 
+    numeric_cols = ['G', 'H', 'S', 'Cp', 'V',
+                  'a1.a', 'a2.b', 'a3.c', 'a4.d',
+                  'c1.e', 'c2.f', 'omega.lambda', 'z.T']
+    
+    df_mod_OBIGT[numeric_cols] = df_mod_OBIGT[numeric_cols].apply(pd.to_numeric)
+    
     return mod_OBIGT(df_mod_OBIGT, messages=messages)
 
 
@@ -2122,8 +2240,20 @@ def mod_OBIGT(*args, messages=True, **kwargs):
     capture.capture_r_output()
         
     if isinstance(args[0], pd.DataFrame):
+        
+        for col in args[0].columns:
+            if sum(args[0][col].isna()) > 0 and args[0][col].dtype == "O":
+                if messages:
+                    print("mod_OBIGT Warning: The column " + str(col) + " "
+                          "has dtype 'object' but contains NA values that "
+                          "can be interpreted as 'float'. Removing NA values.")
+                args[0][col] = args[0][col].fillna('')
+        
+        
         arg_list = list(args)
+        
         arg_list[0] = ro.conversion.py2rpy(arg_list[0])
+        
         args = tuple(arg_list)
     else:
         pass
